#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from dataloader.io.image_loader import LoadDictImage
from dataloader.io.image_reader import Cv2ImageReader
from dataloader.io.loader import Loader
from dataloader.io.reader import Reader
from dataloader.io.registry import reader_registry
